#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from airbyte_cdk.sources.file_based.stream.concurrent.cursor import FileBasedConcurrentCursor
from airbyte_cdk.test.state_builder import StateBuilder
from unit_tests.sources.file_based.helpers import LowHistoryLimitConcurrentCursor
from unit_tests.sources.file_based.scenarios.file_based_source_builder import FileBasedSourceBuilder
from unit_tests.sources.file_based.scenarios.scenario_builder import IncrementalScenarioConfig, TestScenarioBuilder

single_csv_input_state_is_earlier_scenario_concurrent = (
    TestScenarioBuilder()
    .set_name("single_csv_input_state_is_earlier_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11", "val12"),
                        ("val21", "val22"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                }
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {"some_old_file.csv": "2023-06-01T03:54:07.000000Z"},
                    "_ab_source_file_last_modified": "2023-06-01T03:54:07.000000Z_some_old_file.csv",
                },
            )
            .build(),
        )
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11",
                    "col2": "val12",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21",
                    "col2": "val22",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {"some_old_file.csv": "2023-06-01T03:54:07.000000Z", "a.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_a.csv",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                }
            ]
        }
    )
).build()

single_csv_file_is_skipped_if_same_modified_at_as_in_history_concurrent = (
    TestScenarioBuilder()
    .set_name("single_csv_file_is_skipped_if_same_modified_at_as_in_history_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11", "val12"),
                        ("val21", "val22"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                }
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {"a.csv": "2023-06-05T03:54:07.000000Z"},
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_a.csv",
                },
            )
            .build(),
        )
    )
    .set_expected_records(
        [
            {
                "history": {"a.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_a.csv",
            }
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                }
            ]
        }
    )
).build()

single_csv_file_is_synced_if_modified_at_is_more_recent_than_in_history_concurrent = (
    TestScenarioBuilder()
    .set_name("single_csv_file_is_synced_if_modified_at_is_more_recent_than_in_history_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11", "val12"),
                        ("val21", "val22"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000Z",
                }
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {"a.csv": "2023-06-01T03:54:07.000000Z"},
                    "_ab_source_file_last_modified": "2023-06-01T03:54:07.000000Z_a.csv",
                },
            )
            .build(),
        )
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11",
                    "col2": "val12",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21",
                    "col2": "val22",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {"a.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_a.csv",
            },
        ]
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                }
            ]
        }
    )
).build()

single_csv_no_input_state_scenario_concurrent = (
    TestScenarioBuilder()
    .set_name("single_csv_input_state_is_earlier_again_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11", "val12"),
                        ("val21", "val22"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                }
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11",
                    "col2": "val12",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21",
                    "col2": "val22",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {"a.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_a.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
).build()

multi_csv_same_timestamp_scenario_concurrent = (
    TestScenarioBuilder()
    .set_name("multi_csv_same_timestamp_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {"a.csv": "2023-06-05T03:54:07.000000Z", "b.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_b.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
).build()

single_csv_input_state_is_later_scenario_concurrent = (
    TestScenarioBuilder()
    .set_name("single_csv_input_state_is_later_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11", "val12"),
                        ("val21", "val22"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                }
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11",
                    "col2": "val12",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21",
                    "col2": "val22",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "recent_file.csv": "2023-07-15T23:59:59.000000Z",
                    "a.csv": "2023-06-05T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-07-15T23:59:59.000000Z_recent_file.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {"recent_file.csv": "2023-07-15T23:59:59.000000Z"},
                    "_ab_source_file_last_modified": "2023-07-15T23:59:59.000000Z_recent_file.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_different_timestamps_scenario_concurrent = (
    TestScenarioBuilder()
    .set_name("multi_csv_stream_different_timestamps_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-04T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-04T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-04T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "a.csv": "2023-06-04T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-04T03:54:07.000000Z_a.csv",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {"a.csv": "2023-06-04T03:54:07.000000Z", "b.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_b.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
).build()

multi_csv_per_timestamp_scenario_concurrent = (
    TestScenarioBuilder()
    .set_name("multi_csv_per_timestamp_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {"a.csv": "2023-06-05T03:54:07.000000Z", "b.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_b.csv",
            },
            {
                "data": {
                    "col1": "val11c",
                    "col2": "val12c",
                    "col3": "val13c",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21c",
                    "col2": "val22c",
                    "col3": "val23c",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "a.csv": "2023-06-05T03:54:07.000000Z",
                    "b.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-06T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_c.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
).build()

multi_csv_skip_file_if_already_in_history_concurrent = (
    TestScenarioBuilder()
    .set_name("skip_files_already_in_history_concurrent")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            # {"data": {"col1": "val11a", "col2": "val12a"}, "stream": "stream1"}, # this file is skipped
            # {"data": {"col1": "val21a", "col2": "val22a"}, "stream": "stream1"}, # this file is skipped
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {"a.csv": "2023-06-05T03:54:07.000000Z", "b.csv": "2023-06-05T03:54:07.000000Z"},
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_b.csv",
            },
            {
                "data": {
                    "col1": "val11c",
                    "col2": "val12c",
                    "col3": "val13c",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21c",
                    "col2": "val22c",
                    "col3": "val23c",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "a.csv": "2023-06-05T03:54:07.000000Z",
                    "b.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-06T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_c.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {"history": {"a.csv": "2023-06-05T03:54:07.000000Z"}, "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_a.csv"},
            )
            .build(),
        )
    )
).build()

multi_csv_include_missing_files_within_history_range_concurrent_cursor_is_newer = (
    TestScenarioBuilder()
    .set_name("multi_csv_include_missing_files_within_history_range_concurrent_cursor_is_newer")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            # {"data": {"col1": "val11a", "col2": "val12a"}, "stream": "stream1"}, # this file is skipped
            # {"data": {"col1": "val21a", "col2": "val22a"}, "stream": "stream1"}, # this file is skipped
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val11c", "col2": "val12c", "col3": "val13c"}, "stream": "stream1"}, # this file is skipped
            # {"data": {"col1": "val21c", "col2": "val22c", "col3": "val23c"}, "stream": "stream1"}, # this file is skipped
            {
                "history": {
                    "a.csv": "2023-06-05T03:54:07.000000Z",
                    "b.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-06T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_c.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {"a.csv": "2023-06-05T03:54:07.000000Z", "c.csv": "2023-06-06T03:54:07.000000Z"},
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_c.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_include_missing_files_within_history_range_concurrent_cursor_is_older = (
    TestScenarioBuilder()
    .set_name("multi_csv_include_missing_files_within_history_range_concurrent_cursor_is_older")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(FileBasedConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            # {"data": {"col1": "val11a", "col2": "val12a"}, "stream": "stream1"}, # this file is skipped
            # {"data": {"col1": "val21a", "col2": "val22a"}, "stream": "stream1"}, # this file is skipped
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            # {"data": {"col1": "val11c", "col2": "val12c", "col3": "val13c"}, "stream": "stream1"}, # this file is skipped
            # {"data": {"col1": "val21c", "col2": "val22c", "col3": "val23c"}, "stream": "stream1"}, # this file is skipped
            {
                "history": {
                    "a.csv": "2023-06-05T03:54:07.000000Z",
                    "b.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-06T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_c.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {"a.csv": "2023-06-05T03:54:07.000000Z", "c.csv": "2023-06-06T03:54:07.000000Z"},
                    "_ab_source_file_last_modified": "2023-06-03T03:54:07.000000Z_x.csv",
                },
            )
            .build()
        )
    )
).build()

multi_csv_remove_old_files_if_history_is_full_scenario_concurrent_cursor_is_newer = (
    TestScenarioBuilder()
    .set_name("multi_csv_remove_old_files_if_history_is_full_scenario_concurrent_cursor_is_newer")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-07T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-10T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "very_old_file.csv": "2023-06-02T03:54:07.000000Z",
                    "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    "a.csv": "2023-06-06T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_old_file_same_timestamp_as_a.csv",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-07T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-07T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    "a.csv": "2023-06-06T03:54:07.000000Z",
                    "b.csv": "2023-06-07T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-07T03:54:07.000000Z_b.csv",
            },
            {
                "data": {
                    "col1": "val11c",
                    "col2": "val12c",
                    "col3": "val13c",
                    "_ab_source_file_last_modified": "2023-06-10T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21c",
                    "col2": "val22c",
                    "col3": "val23c",
                    "_ab_source_file_last_modified": "2023-06-10T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    "b.csv": "2023-06-07T03:54:07.000000Z",
                    "c.csv": "2023-06-10T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-10T03:54:07.000000Z_c.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "very_very_old_file.csv": "2023-06-01T03:54:07.000000Z",
                        "very_old_file.csv": "2023-06-02T03:54:07.000000Z",
                        "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_old_file_same_timestamp_as_a.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_remove_old_files_if_history_is_full_scenario_concurrent_cursor_is_older = (
    TestScenarioBuilder()
    .set_name("multi_csv_remove_old_files_if_history_is_full_scenario_concurrent_cursor_is_older")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-07T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-10T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "very_old_file.csv": "2023-06-02T03:54:07.000000Z",
                    "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    "a.csv": "2023-06-06T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z_old_file_same_timestamp_as_a.csv",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-07T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-07T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    "a.csv": "2023-06-06T03:54:07.000000Z",
                    "b.csv": "2023-06-07T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-07T03:54:07.000000Z_b.csv",
            },
            {
                "data": {
                    "col1": "val11c",
                    "col2": "val12c",
                    "col3": "val13c",
                    "_ab_source_file_last_modified": "2023-06-10T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21c",
                    "col2": "val22c",
                    "col3": "val23c",
                    "_ab_source_file_last_modified": "2023-06-10T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    "b.csv": "2023-06-07T03:54:07.000000Z",
                    "c.csv": "2023-06-10T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-10T03:54:07.000000Z_c.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "very_very_old_file.csv": "2023-06-01T03:54:07.000000Z",
                        "very_old_file.csv": "2023-06-02T03:54:07.000000Z",
                        "old_file_same_timestamp_as_a.csv": "2023-06-06T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-05-01T03:54:07.000000Z_very_very_very_old_file.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_same_timestamp_more_files_than_history_size_scenario_concurrent_cursor_is_newer = (
    TestScenarioBuilder()
    .set_name("multi_csv_same_timestamp_more_files_than_history_size_scenario_concurrent_cursor_is_newer")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11c",
                    "col2": "val12c",
                    "col3": "val13c",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21c",
                    "col2": "val22c",
                    "col3": "val23c",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11d",
                    "col2": "val12d",
                    "col3": "val13d",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "d.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21d",
                    "col2": "val22d",
                    "col3": "val23d",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "d.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "b.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-05T03:54:07.000000Z",
                    "d.csv": "2023-06-05T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_d.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
).build()

multi_csv_same_timestamp_more_files_than_history_size_scenario_concurrent_cursor_is_older = (
    TestScenarioBuilder()
    .set_name("multi_csv_same_timestamp_more_files_than_history_size_scenario_concurrent_cursor_is_older")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11c",
                    "col2": "val12c",
                    "col3": "val13c",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21c",
                    "col2": "val22c",
                    "col3": "val23c",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "c.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val11d",
                    "col2": "val12d",
                    "col3": "val13d",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "d.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21d",
                    "col2": "val22d",
                    "col3": "val23d",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "d.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "b.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-05T03:54:07.000000Z",
                    "d.csv": "2023-06-05T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_d.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=[],
        )
    )
).build()

multi_csv_sync_recent_files_if_history_is_incomplete_scenario_concurrent_cursor_is_older = (
    TestScenarioBuilder()
    .set_name("multi_csv_sync_recent_files_if_history_is_incomplete_scenario_concurrent_cursor_is_older")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
            }
        )
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
        .set_file_type("csv")
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "history": {
                    "b.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-05T03:54:07.000000Z",
                    "d.csv": "2023-06-05T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_d.csv",
            }
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "b.csv": "2023-06-05T03:54:07.000000Z",
                        "c.csv": "2023-06-05T03:54:07.000000Z",
                        "d.csv": "2023-06-05T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_b.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_sync_recent_files_if_history_is_incomplete_scenario_concurrent_cursor_is_newer = (
    TestScenarioBuilder()
    .set_name("multi_csv_sync_recent_files_if_history_is_incomplete_scenario_concurrent_cursor_is_newer")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
            }
        )
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
        .set_file_type("csv")
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                    "history": {
                        "b.csv": "2023-06-05T03:54:07.000000Z",
                        "c.csv": "2023-06-05T03:54:07.000000Z",
                        "d.csv": "2023-06-05T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_d.csv",
            }
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "b.csv": "2023-06-05T03:54:07.000000Z",
                        "c.csv": "2023-06-05T03:54:07.000000Z",
                        "d.csv": "2023-06-05T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z_d.csv",
                },
            )
            .build(),
        )
    )
).build()


multi_csv_sync_files_within_time_window_if_history_is_incomplete__different_timestamps_scenario_concurrent_cursor_is_older = (
    TestScenarioBuilder()
    .set_name("multi_csv_sync_files_within_time_window_if_history_is_incomplete__different_timestamps_scenario_concurrent_cursor_is_older")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-07T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-08T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            # {"data": {"col1": "val11a", "col2": "val12a"}, "stream": "stream1"}, # This file is skipped because it is older than the time_window
            # {"data": {"col1": "val21a", "col2": "val22a"}, "stream": "stream1"},
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "c.csv": "2023-06-07T03:54:07.000000Z",
                    "d.csv": "2023-06-08T03:54:07.000000Z",
                    "e.csv": "2023-06-08T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_e.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "c.csv": "2023-06-07T03:54:07.000000Z",
                        "d.csv": "2023-06-08T03:54:07.000000Z",
                        "e.csv": "2023-06-08T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_e.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_sync_files_within_time_window_if_history_is_incomplete__different_timestamps_scenario_concurrent_cursor_is_newer = (
    TestScenarioBuilder()
    .set_name("multi_csv_sync_files_within_time_window_if_history_is_incomplete__different_timestamps_scenario_concurrent_cursor_is_newer")
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-07T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-08T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            # {"data": {"col1": "val11a", "col2": "val12a"}, "stream": "stream1"}, # This file is skipped because it is older than the time_window
            # {"data": {"col1": "val21a", "col2": "val22a"}, "stream": "stream1"},
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "c.csv": "2023-06-07T03:54:07.000000Z",
                    "d.csv": "2023-06-08T03:54:07.000000Z",
                    "e.csv": "2023-06-08T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_e.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "c.csv": "2023-06-07T03:54:07.000000Z",
                        "d.csv": "2023-06-08T03:54:07.000000Z",
                        "e.csv": "2023-06-08T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_e.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_sync_files_within_history_time_window_if_history_is_incomplete_different_timestamps_scenario_concurrent_cursor_is_newer = (
    TestScenarioBuilder()
    .set_name(
        "multi_csv_sync_files_within_history_time_window_if_history_is_incomplete_different_timestamps_scenario_concurrent_cursor_is_newer"
    )
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-07T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-08T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "a.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-07T03:54:07.000000Z",
                    "d.csv": "2023-06-08T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_d.csv",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "b.csv": "2023-06-06T03:54:07.000000Z",
                    "c.csv": "2023-06-07T03:54:07.000000Z",
                    "d.csv": "2023-06-08T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_d.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "old_file.csv": "2023-06-05T00:00:00.000000Z",
                        "c.csv": "2023-06-07T03:54:07.000000Z",
                        "d.csv": "2023-06-08T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_d.csv",
                },
            )
            .build(),
        )
    )
).build()

multi_csv_sync_files_within_history_time_window_if_history_is_incomplete_different_timestamps_scenario_concurrent_cursor_is_older = (
    TestScenarioBuilder()
    .set_name(
        "multi_csv_sync_files_within_history_time_window_if_history_is_incomplete_different_timestamps_scenario_concurrent_cursor_is_older"
    )
    .set_config(
        {
            "streams": [
                {
                    "name": "stream1",
                    "format": {"filetype": "csv"},
                    "globs": ["*.csv"],
                    "validation_policy": "Emit Record",
                    "days_to_sync_if_history_is_full": 3,
                }
            ]
        }
    )
    .set_source_builder(
        FileBasedSourceBuilder()
        .set_files(
            {
                "a.csv": {
                    "contents": [
                        ("col1", "col2"),
                        ("val11a", "val12a"),
                        ("val21a", "val22a"),
                    ],
                    "last_modified": "2023-06-05T03:54:07.000000Z",
                },
                "b.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11b", "val12b", "val13b"),
                        ("val21b", "val22b", "val23b"),
                    ],
                    "last_modified": "2023-06-06T03:54:07.000000Z",
                },
                "c.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11c", "val12c", "val13c"),
                        ("val21c", "val22c", "val23c"),
                    ],
                    "last_modified": "2023-06-07T03:54:07.000000Z",
                },
                "d.csv": {
                    "contents": [
                        ("col1", "col2", "col3"),
                        ("val11d", "val12d", "val13d"),
                        ("val21d", "val22d", "val23d"),
                    ],
                    "last_modified": "2023-06-08T03:54:07.000000Z",
                },
            }
        )
        .set_file_type("csv")
        .set_cursor_cls(LowHistoryLimitConcurrentCursor)
    )
    .set_expected_catalog(
        {
            "streams": [
                {
                    "default_cursor_field": ["_ab_source_file_last_modified"],
                    "json_schema": {
                        "type": "object",
                        "properties": {
                            "col1": {
                                "type": ["null", "string"],
                            },
                            "col2": {
                                "type": ["null", "string"],
                            },
                            "col3": {
                                "type": ["null", "string"],
                            },
                            "_ab_source_file_last_modified": {"type": "string"},
                            "_ab_source_file_url": {"type": "string"},
                        },
                    },
                    "name": "stream1",
                    "source_defined_cursor": True,
                    "supported_sync_modes": ["full_refresh", "incremental"],
                }
            ]
        }
    )
    .set_expected_records(
        [
            {
                "data": {
                    "col1": "val11a",
                    "col2": "val12a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21a",
                    "col2": "val22a",
                    "_ab_source_file_last_modified": "2023-06-05T03:54:07.000000Z",
                    "_ab_source_file_url": "a.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "a.csv": "2023-06-05T03:54:07.000000Z",
                    "c.csv": "2023-06-07T03:54:07.000000Z",
                    "d.csv": "2023-06-08T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_d.csv",
            },
            {
                "data": {
                    "col1": "val11b",
                    "col2": "val12b",
                    "col3": "val13b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "data": {
                    "col1": "val21b",
                    "col2": "val22b",
                    "col3": "val23b",
                    "_ab_source_file_last_modified": "2023-06-06T03:54:07.000000Z",
                    "_ab_source_file_url": "b.csv",
                },
                "stream": "stream1",
            },
            {
                "history": {
                    "b.csv": "2023-06-06T03:54:07.000000Z",
                    "c.csv": "2023-06-07T03:54:07.000000Z",
                    "d.csv": "2023-06-08T03:54:07.000000Z",
                },
                "_ab_source_file_last_modified": "2023-06-08T03:54:07.000000Z_d.csv",
            },
        ]
    )
    .set_incremental_scenario_config(
        IncrementalScenarioConfig(
            input_state=StateBuilder()
            .with_stream_state(
                "stream1",
                {
                    "history": {
                        "old_file.csv": "2023-06-05T00:00:00.000000Z",
                        "c.csv": "2023-06-07T03:54:07.000000Z",
                        "d.csv": "2023-06-08T03:54:07.000000Z",
                    },
                    "_ab_source_file_last_modified": "2023-06-04T00:00:00.000000Z_very_old_file.csv",
                },
            )
            .build(),
        )
    )
).build()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import functools
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, List, Mapping, MutableMapping, Optional, Protocol, Tuple

from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
from airbyte_cdk.sources.message import MessageRepository
from airbyte_cdk.sources.streams import FULL_REFRESH_SENTINEL_STATE_KEY
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.record import Record
from airbyte_cdk.sources.streams.concurrent.state_converters.abstract_stream_state_converter import AbstractStreamStateConverter


def _extract_value(mapping: Mapping[str, Any], path: List[str]) -> Any:
    return functools.reduce(lambda a, b: a[b], path, mapping)


class Comparable(Protocol):
    """Protocol for annotating comparable types."""

    @abstractmethod
    def __lt__(self: "Comparable", other: "Comparable") -> bool:
        pass


class CursorField:
    def __init__(self, cursor_field_key: str) -> None:
        self.cursor_field_key = cursor_field_key

    def extract_value(self, record: Record) -> Comparable:
        cursor_value = record.data.get(self.cursor_field_key)
        if cursor_value is None:
            raise ValueError(f"Could not find cursor field {self.cursor_field_key} in record")
        return cursor_value  # type: ignore  # we assume that the value the path points at is a comparable


class Cursor(ABC):
    @property
    @abstractmethod
    def state(self) -> MutableMapping[str, Any]:
        ...

    @abstractmethod
    def observe(self, record: Record) -> None:
        """
        Indicate to the cursor that the record has been emitted
        """
        raise NotImplementedError()

    @abstractmethod
    def close_partition(self, partition: Partition) -> None:
        """
        Indicate to the cursor that the partition has been successfully processed
        """
        raise NotImplementedError()

    @abstractmethod
    def ensure_at_least_one_state_emitted(self) -> None:
        """
        State messages are emitted when a partition is closed. However, the platform expects at least one state to be emitted per sync per
        stream. Hence, if no partitions are generated, this method needs to be called.
        """
        raise NotImplementedError()


class FinalStateCursor(Cursor):
    """Cursor that is used to guarantee at least one state message is emitted for a concurrent stream."""

    def __init__(
        self,
        stream_name: str,
        stream_namespace: Optional[str],
        message_repository: MessageRepository,
    ) -> None:
        self._stream_name = stream_name
        self._stream_namespace = stream_namespace
        self._message_repository = message_repository
        # Normally the connector state manager operates at the source-level. However, we only need it to write the sentinel
        # state message rather than manage overall source state. This is also only temporary as we move to the resumable
        # full refresh world where every stream uses a FileBasedConcurrentCursor with incremental state.
        self._connector_state_manager = ConnectorStateManager(stream_instance_map={})
        self._has_closed_at_least_one_slice = False

    @property
    def state(self) -> MutableMapping[str, Any]:
        return {FULL_REFRESH_SENTINEL_STATE_KEY: True}

    def observe(self, record: Record) -> None:
        pass

    def close_partition(self, partition: Partition) -> None:
        pass

    def ensure_at_least_one_state_emitted(self) -> None:
        """
        Used primarily for full refresh syncs that do not have a valid cursor value to emit at the end of a sync
        """

        self._connector_state_manager.update_state_for_stream(self._stream_name, self._stream_namespace, self.state)
        state_message = self._connector_state_manager.create_state_message(self._stream_name, self._stream_namespace)
        self._message_repository.emit_message(state_message)


class ConcurrentCursor(Cursor):
    _START_BOUNDARY = 0
    _END_BOUNDARY = 1

    def __init__(
        self,
        stream_name: str,
        stream_namespace: Optional[str],
        stream_state: Any,
        message_repository: MessageRepository,
        connector_state_manager: ConnectorStateManager,
        connector_state_converter: AbstractStreamStateConverter,
        cursor_field: CursorField,
        slice_boundary_fields: Optional[Tuple[str, str]],
        start: Optional[Any],
    ) -> None:
        self._stream_name = stream_name
        self._stream_namespace = stream_namespace
        self._message_repository = message_repository
        self._connector_state_converter = connector_state_converter
        self._connector_state_manager = connector_state_manager
        self._cursor_field = cursor_field
        # To see some example where the slice boundaries might not be defined, check https://github.com/airbytehq/airbyte/blob/1ce84d6396e446e1ac2377362446e3fb94509461/airbyte-integrations/connectors/source-stripe/source_stripe/streams.py#L363-L379
        self._slice_boundary_fields = slice_boundary_fields if slice_boundary_fields else tuple()
        self._start = start
        self._most_recent_record: Optional[Record] = None
        self._has_closed_at_least_one_slice = False
        self.start, self._concurrent_state = self._get_concurrent_state(stream_state)

    @property
    def state(self) -> MutableMapping[str, Any]:
        return self._concurrent_state

    def _get_concurrent_state(self, state: MutableMapping[str, Any]) -> Tuple[datetime, MutableMapping[str, Any]]:
        if self._connector_state_converter.is_state_message_compatible(state):
            return self._start or self._connector_state_converter.zero_value, self._connector_state_converter.deserialize(state)
        return self._connector_state_converter.convert_from_sequential_state(self._cursor_field, state, self._start)

    def observe(self, record: Record) -> None:
        if self._slice_boundary_fields:
            # Given that slicing is done using the cursor field, we don't need to observe the record as we assume slices will describe what
            # has been emitted. Assuming there is a chance that records might not be yet populated for the most recent slice, use a lookback
            # window
            return

        if not self._most_recent_record or self._extract_cursor_value(self._most_recent_record) < self._extract_cursor_value(record):
            self._most_recent_record = record

    def _extract_cursor_value(self, record: Record) -> Any:
        return self._connector_state_converter.parse_value(self._cursor_field.extract_value(record))

    def close_partition(self, partition: Partition) -> None:
        slice_count_before = len(self.state.get("slices", []))
        self._add_slice_to_state(partition)
        if slice_count_before < len(self.state["slices"]):  # only emit if at least one slice has been processed
            self._merge_partitions()
            self._emit_state_message()
        self._has_closed_at_least_one_slice = True

    def _add_slice_to_state(self, partition: Partition) -> None:
        if self._slice_boundary_fields:
            if "slices" not in self.state:
                raise RuntimeError(
                    f"The state for stream {self._stream_name} should have at least one slice to delineate the sync start time, but no slices are present. This is unexpected. Please contact Support."
                )
            self.state["slices"].append(
                {
                    "start": self._extract_from_slice(partition, self._slice_boundary_fields[self._START_BOUNDARY]),
                    "end": self._extract_from_slice(partition, self._slice_boundary_fields[self._END_BOUNDARY]),
                }
            )
        elif self._most_recent_record:
            if self._has_closed_at_least_one_slice:
                # If we track state value using records cursor field, we can only do that if there is one partition. This is because we save
                # the state every time we close a partition. We assume that if there are multiple slices, they need to be providing
                # boundaries. There are cases where partitions could not have boundaries:
                # * The cursor should be per-partition
                # * The stream state is actually the parent stream state
                # There might be other cases not listed above. Those are not supported today hence the stream should not use this cursor for
                # state management. For the specific user that was affected with this issue, we need to:
                # * Fix state tracking (which is currently broken)
                # * Make the new version available
                # * (Probably) ask the user to reset the stream to avoid data loss
                raise ValueError(
                    "Given that slice_boundary_fields is not defined and that per-partition state is not supported, only one slice is "
                    "expected. Please contact the Airbyte team."
                )

            self.state["slices"].append(
                {
                    self._connector_state_converter.START_KEY: self.start,
                    self._connector_state_converter.END_KEY: self._extract_cursor_value(self._most_recent_record),
                }
            )

    def _emit_state_message(self) -> None:
        self._connector_state_manager.update_state_for_stream(
            self._stream_name,
            self._stream_namespace,
            self._connector_state_converter.convert_to_sequential_state(self._cursor_field, self.state),
        )
        # TODO: if we migrate stored state to the concurrent state format
        #  (aka stop calling self._connector_state_converter.convert_to_sequential_state`), we'll need to cast datetimes to string or
        #  int before emitting state
        state_message = self._connector_state_manager.create_state_message(self._stream_name, self._stream_namespace)
        self._message_repository.emit_message(state_message)

    def _merge_partitions(self) -> None:
        self.state["slices"] = self._connector_state_converter.merge_intervals(self.state["slices"])

    def _extract_from_slice(self, partition: Partition, key: str) -> Comparable:
        try:
            _slice = partition.to_slice()
            if not _slice:
                raise KeyError(f"Could not find key `{key}` in empty slice")
            return self._connector_state_converter.parse_value(_slice[key])  # type: ignore  # we expect the devs to specify a key that would return a Comparable
        except KeyError as exception:
            raise KeyError(f"Partition is expected to have key `{key}` but could not be found") from exception

    def ensure_at_least_one_state_emitted(self) -> None:
        """
        The platform expect to have at least one state message on successful syncs. Hence, whatever happens, we expect this method to be
        called.
        """
        self._emit_state_message()



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from unittest.mock import patch

from airbyte_cdk.logger import init_logger
from airbyte_cdk.models import ConfiguredAirbyteCatalog, SyncMode, Type
from airbyte_cdk.sources.declarative.incremental.per_partition_cursor import StreamSlice
from airbyte_cdk.sources.declarative.manifest_declarative_source import ManifestDeclarativeSource
from airbyte_cdk.sources.declarative.retrievers.simple_retriever import SimpleRetriever
from airbyte_cdk.sources.declarative.types import Record

CURSOR_FIELD = "cursor_field"
SYNC_MODE = SyncMode.incremental


class ManifestBuilder:
    def __init__(self):
        self._incremental_sync = {}
        self._partition_router = {}
        self._substream_partition_router = {}

    def with_list_partition_router(self, stream_name, cursor_field, partitions):
        self._partition_router[stream_name] = {
            "type": "ListPartitionRouter",
            "cursor_field": cursor_field,
            "values": partitions,
        }
        return self

    def with_substream_partition_router(self, stream_name):
        self._substream_partition_router[stream_name] = {
            "type": "SubstreamPartitionRouter",
            "parent_stream_configs": [
                {
                    "type": "ParentStreamConfig",
                    "stream": "#/definitions/Rates",
                    "parent_key": "id",
                    "partition_field": "parent_id",

                }
            ]
        }
        return self

    def with_incremental_sync(self, stream_name, start_datetime, end_datetime, datetime_format, cursor_field, step, cursor_granularity):
        self._incremental_sync[stream_name] = {
            "type": "DatetimeBasedCursor",
            "start_datetime": start_datetime,
            "end_datetime": end_datetime,
            "datetime_format": datetime_format,
            "cursor_field": cursor_field,
            "step": step,
            "cursor_granularity": cursor_granularity,
        }
        return self

    def build(self):
        manifest = {
            "version": "0.34.2",
            "type": "DeclarativeSource",
            "check": {"type": "CheckStream", "stream_names": ["Rates"]},
            "definitions": {
                "AnotherStream": {
                    "type": "DeclarativeStream",
                    "name": "AnotherStream",
                    "primary_key": [],
                    "schema_loader": {
                        "type": "InlineSchemaLoader",
                        "schema": {"$schema": "http://json-schema.org/schema#", "properties": {"id": {"type": "string"}}, "type": "object"},
                    },
                    "retriever": {
                        "type": "SimpleRetriever",
                        "requester": {
                            "type": "HttpRequester",
                            "url_base": "https://api.apilayer.com",
                            "path": "/exchangerates_data/latest",
                            "http_method": "GET",
                        },
                        "record_selector": {"type": "RecordSelector", "extractor": {"type": "DpathExtractor", "field_path": []}},
                    },
                },
                "Rates": {
                    "type": "DeclarativeStream",
                    "name": "Rates",
                    "primary_key": [],
                    "schema_loader": {
                        "type": "InlineSchemaLoader",
                        "schema": {"$schema": "http://json-schema.org/schema#", "properties": {}, "type": "object"},
                    },
                    "retriever": {
                        "type": "SimpleRetriever",
                        "requester": {
                            "type": "HttpRequester",
                            "url_base": "https://api.apilayer.com",
                            "path": "/exchangerates_data/latest",
                            "http_method": "GET",
                        },
                        "record_selector": {"type": "RecordSelector", "extractor": {"type": "DpathExtractor", "field_path": []}},
                    },
                },
            },
            "streams": [
                {"$ref": "#/definitions/Rates"},
                {"$ref": "#/definitions/AnotherStream"}
            ],
            "spec": {
                "connection_specification": {
                    "$schema": "http://json-schema.org/draft-07/schema#",
                    "type": "object",
                    "required": [],
                    "properties": {},
                    "additionalProperties": True,
                },
                "documentation_url": "https://example.org",
                "type": "Spec",
            },
        }
        for stream_name, incremental_sync_definition in self._incremental_sync.items():
            manifest["definitions"][stream_name]["incremental_sync"] = incremental_sync_definition
        for stream_name, partition_router_definition in self._partition_router.items():
            manifest["definitions"][stream_name]["retriever"]["partition_router"] = partition_router_definition
        for stream_name, partition_router_definition in self._substream_partition_router.items():
            manifest["definitions"][stream_name]["retriever"]["partition_router"] = partition_router_definition
        return manifest


def test_given_state_for_only_some_partition_when_stream_slices_then_create_slices_using_state_or_start_from_start_datetime():
    source = ManifestDeclarativeSource(
        source_config=ManifestBuilder()
        .with_list_partition_router("Rates", "partition_field", ["1", "2"])
        .with_incremental_sync(
            "Rates",
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .build()
    )
    stream_instance = source.streams({})[0]
    stream_instance.state = {
        "states": [
            {
                "partition": {"partition_field": "1"},
                "cursor": {CURSOR_FIELD: "2022-02-01"},
            }
        ]
    }

    slices = stream_instance.stream_slices(
        sync_mode=SYNC_MODE,
        stream_state={},
    )

    assert list(slices) == [
        {"partition_field": "1", "start_time": "2022-02-01", "end_time": "2022-02-28"},
        {"partition_field": "2", "start_time": "2022-01-01", "end_time": "2022-01-31"},
        {"partition_field": "2", "start_time": "2022-02-01", "end_time": "2022-02-28"},
    ]


def test_given_record_for_partition_when_read_then_update_state():
    source = ManifestDeclarativeSource(
        source_config=ManifestBuilder()
        .with_list_partition_router("Rates", "partition_field", ["1", "2"])
        .with_incremental_sync(
            "Rates",
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .build()
    )
    stream_instance = source.streams({})[0]
    list(stream_instance.stream_slices(sync_mode=SYNC_MODE))

    stream_slice = StreamSlice(partition={"partition_field": "1"},
                               cursor_slice={"start_time": "2022-01-01", "end_time": "2022-01-31"})
    with patch.object(
            SimpleRetriever, "_read_pages",
            side_effect=[[Record({"a record key": "a record value", CURSOR_FIELD: "2022-01-15"}, stream_slice)]]
    ):
        list(
            stream_instance.read_records(
                sync_mode=SYNC_MODE,
                stream_slice=stream_slice,
                stream_state={"states": []},
                cursor_field=CURSOR_FIELD,
            )
        )

    assert stream_instance.state == {
        "states": [
            {
                "partition": {"partition_field": "1"},
                "cursor": {CURSOR_FIELD: "2022-01-15"},
            }
        ]
    }


def test_substream_without_input_state():
    test_source = ManifestDeclarativeSource(
        source_config=ManifestBuilder()
        .with_substream_partition_router("AnotherStream")
        .with_incremental_sync(
            "Rates",
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .with_incremental_sync(
            "AnotherStream",
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .build()
    )

    stream_instance = test_source.streams({})[1]

    stream_slice = StreamSlice(partition={"parent_id": "1"},
                               cursor_slice={"start_time": "2022-01-01", "end_time": "2022-01-31"})

    with patch.object(
            SimpleRetriever, "_read_pages", side_effect=[[Record({"id": "1", CURSOR_FIELD: "2022-01-15"}, stream_slice)],
                                                         [Record({"id": "2", CURSOR_FIELD: "2022-01-15"}, stream_slice)]]
    ):
        slices = list(stream_instance.stream_slices(sync_mode=SYNC_MODE))
        assert list(slices) == [
            StreamSlice(partition={"parent_id": "1", "parent_slice": {}, },
                        cursor_slice={"start_time": "2022-01-01", "end_time": "2022-01-31"}),
            StreamSlice(partition={"parent_id": "1", "parent_slice": {}, },
                        cursor_slice={"start_time": "2022-02-01", "end_time": "2022-02-28"}),
            StreamSlice(partition={"parent_id": "2", "parent_slice": {}, },
                        cursor_slice={"start_time": "2022-01-01", "end_time": "2022-01-31"}),
            StreamSlice(partition={"parent_id": "2", "parent_slice": {}, },
                        cursor_slice={"start_time": "2022-02-01", "end_time": "2022-02-28"}),
        ]


def test_substream_with_legacy_input_state():
    source = ManifestDeclarativeSource(
        source_config=ManifestBuilder()
        .with_substream_partition_router("AnotherStream")
        .with_incremental_sync(
            "Rates",
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .with_incremental_sync(
            "AnotherStream",
            start_datetime="2022-01-01",
            end_datetime="2022-02-28",
            datetime_format="%Y-%m-%d",
            cursor_field=CURSOR_FIELD,
            step="P1M",
            cursor_granularity="P1D",
        )
        .build()
    )

    stream_instance = source.streams({})[1]

    input_state = {
        "states": [
            {
                "partition": {"item_id": "an_item_id",
                              "parent_slice": {"end_time": "1629640663", "start_time": "1626962264"},
                              },
                "cursor": {
                    "updated_at": "1709058818"
                }
            }
        ]
    }
    stream_instance.state = input_state

    stream_slice = StreamSlice(partition={"parent_id": "1"},
                               cursor_slice={"start_time": "2022-01-01", "end_time": "2022-01-31"})

    logger = init_logger("airbyte")
    configured_catalog = ConfiguredAirbyteCatalog(
        streams=[
            {
                "stream": {"name": "AnotherStream", "json_schema": {}, "supported_sync_modes": ["incremental"]},
                "sync_mode": "incremental",
                "destination_sync_mode": "overwrite",
            },
        ]
    )

    with patch.object(
            SimpleRetriever, "_read_pages", side_effect=[
                [Record({"id": "1", CURSOR_FIELD: "2022-01-15"}, stream_slice)],
                [Record({"parent_id": "1", CURSOR_FIELD: "2022-01-15"}, stream_slice)],
                [Record({"id": "2", CURSOR_FIELD: "2022-01-15"}, stream_slice)],
                [Record({"parent_id": "2", CURSOR_FIELD: "2022-01-15"}, stream_slice)]
            ]
    ):
        messages = list(source.read(logger, {}, configured_catalog, input_state))

        output_state_message = [message for message in messages if message.type == Type.STATE][0]

        expected_state = {"states": [
            {
                "cursor": {
                    CURSOR_FIELD: "2022-01-15"
                },
                "partition": {"parent_id": "1", "parent_slice": {}}
            }
        ]}

        assert output_state_message.state.stream.stream_state == expected_state



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import unittest
from datetime import datetime
from unittest.mock import Mock, PropertyMock

from airbyte_cdk.sources.file_based.availability_strategy.default_file_based_availability_strategy import (
    DefaultFileBasedAvailabilityStrategy,
)
from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig
from airbyte_cdk.sources.file_based.config.jsonl_format import JsonlFormat
from airbyte_cdk.sources.file_based.exceptions import CustomFileBasedException
from airbyte_cdk.sources.file_based.file_based_stream_reader import AbstractFileBasedStreamReader
from airbyte_cdk.sources.file_based.file_types.file_type_parser import FileTypeParser
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from airbyte_cdk.sources.file_based.stream import AbstractFileBasedStream

_FILE_WITH_UNKNOWN_EXTENSION = RemoteFile(uri="a.unknown_extension", last_modified=datetime.now(), file_type="csv")
_ANY_CONFIG = FileBasedStreamConfig(
    name="config.name",
    file_type="parquet",
    format=JsonlFormat(),
)
_ANY_SCHEMA = {"key": "value"}


class DefaultFileBasedAvailabilityStrategyTest(unittest.TestCase):
    def setUp(self) -> None:
        self._stream_reader = Mock(spec=AbstractFileBasedStreamReader)
        self._strategy = DefaultFileBasedAvailabilityStrategy(self._stream_reader)

        self._parser = Mock(spec=FileTypeParser)
        self._parser.check_config.return_value = (True, None)
        self._stream = Mock(spec=AbstractFileBasedStream)
        self._stream.get_parser.return_value = self._parser
        self._stream.catalog_schema = _ANY_SCHEMA
        self._stream.config = _ANY_CONFIG
        self._stream.validation_policy = PropertyMock(validate_schema_before_sync=False)
        self._stream.stream_reader = self._stream_reader

    def test_given_file_extension_does_not_match_when_check_availability_and_parsability_then_stream_is_still_available(self) -> None:
        """
        Before, we had a validation on the file extension but it turns out that in production, users sometimes have mismatch there. The
        example we've seen was for JSONL parser but the file extension was just `.json`. Note that there we more than one record extracted
        from this stream so it's not just that the file is one JSON object
        """
        self._stream.get_files.return_value = [_FILE_WITH_UNKNOWN_EXTENSION]
        self._parser.parse_records.return_value = [{"a record": 1}]

        is_available, reason = self._strategy.check_availability_and_parsability(self._stream, Mock(), Mock())

        assert is_available

    def test_not_available_given_no_files(self) -> None:
        """
        If no files are returned, then the stream is not available.
        """
        self._stream.get_files.return_value = []

        is_available, reason = self._strategy.check_availability_and_parsability(self._stream, Mock(), Mock())

        assert not is_available
        assert "No files were identified in the stream" in reason

    def test_parse_records_is_not_called_with_parser_max_n_files_for_parsability_set(self) -> None:
        """
        If the stream parser sets parser_max_n_files_for_parsability to 0, then we should not call parse_records on it
        """
        self._parser.parser_max_n_files_for_parsability = 0
        self._stream.get_files.return_value = [_FILE_WITH_UNKNOWN_EXTENSION]

        is_available, reason = self._strategy.check_availability_and_parsability(self._stream, Mock(), Mock())

        assert is_available
        assert not self._parser.parse_records.called
        assert self._stream_reader.open_file.called

    def test_passing_config_check(self) -> None:
        """
        Test if the DefaultFileBasedAvailabilityStrategy correctly handles the check_config method defined on the parser.
        """
        self._parser.check_config.return_value = (False, "Ran into error")
        is_available, error_message = self._strategy.check_availability_and_parsability(self._stream, Mock(), Mock())
        assert not is_available
        assert "Ran into error" in error_message

    def test_catching_and_raising_custom_file_based_exception(self) -> None:
        """
        Test if the DefaultFileBasedAvailabilityStrategy correctly handles the CustomFileBasedException
        by raising a CheckAvailabilityError when the get_files method is called.
        """
        # Mock the get_files method to raise CustomFileBasedException when called
        self._stream.get_files.side_effect = CustomFileBasedException("Custom exception for testing.")

        # Invoke the check_availability_and_parsability method and check if it correctly handles the exception
        is_available, error_message = self._strategy.check_availability_and_parsability(self._stream, Mock(), Mock())
        assert not is_available
        assert "Custom exception for testing." in error_message



from airbyte_cdk.sources.file_based.stream.abstract_file_based_stream import AbstractFileBasedStream
from airbyte_cdk.sources.file_based.stream.default_file_based_stream import DefaultFileBasedStream

__all__ = ["AbstractFileBasedStream", "DefaultFileBasedStream"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import asyncio
import datetime
import math
from typing import Any, Mapping, Union
from unittest.mock import Mock

import pyarrow as pa
import pytest
from airbyte_cdk.sources.file_based.config.csv_format import CsvFormat
from airbyte_cdk.sources.file_based.config.file_based_stream_config import FileBasedStreamConfig, ValidationPolicy
from airbyte_cdk.sources.file_based.config.jsonl_format import JsonlFormat
from airbyte_cdk.sources.file_based.config.parquet_format import ParquetFormat
from airbyte_cdk.sources.file_based.file_types import ParquetParser
from airbyte_cdk.sources.file_based.remote_file import RemoteFile
from pyarrow import Scalar

_default_parquet_format = ParquetFormat()
_decimal_as_float_parquet_format = ParquetFormat(decimal_as_float=True)


@pytest.mark.parametrize(
    "parquet_type, expected_type, parquet_format",
    [
        pytest.param(pa.bool_(), {"type": "boolean"}, _default_parquet_format, id="test_parquet_bool"),
        pytest.param(pa.int8(), {"type": "integer"}, _default_parquet_format, id="test_parquet_int8"),
        pytest.param(pa.int16(), {"type": "integer"}, _default_parquet_format, id="test_parquet_int16"),
        pytest.param(pa.int32(), {"type": "integer"}, _default_parquet_format, id="test_parquet_int32"),
        pytest.param(pa.int64(), {"type": "integer"}, _default_parquet_format, id="test_parquet_int64"),
        pytest.param(pa.uint8(), {"type": "integer"}, _default_parquet_format, id="test_parquet_uint8"),
        pytest.param(pa.uint16(), {"type": "integer"}, _default_parquet_format, id="test_parquet_uint16"),
        pytest.param(pa.uint32(), {"type": "integer"}, _default_parquet_format, id="test_parquet_uint32"),
        pytest.param(pa.uint64(), {"type": "integer"}, _default_parquet_format, id="test_parquet_uint64"),
        pytest.param(pa.float16(), {"type": "number"}, _default_parquet_format, id="test_parquet_float16"),
        pytest.param(pa.float32(), {"type": "number"}, _default_parquet_format, id="test_parquet_float32"),
        pytest.param(pa.float64(), {"type": "number"}, _default_parquet_format, id="test_parquet_float64"),
        pytest.param(pa.time32("s"), {"type": "string"}, _default_parquet_format, id="test_parquet_time32s"),
        pytest.param(pa.time32("ms"), {"type": "string"}, _default_parquet_format, id="test_parquet_time32ms"),
        pytest.param(pa.time64("us"), {"type": "string"}, _default_parquet_format, id="test_parquet_time64us"),
        pytest.param(pa.time64("ns"), {"type": "string"}, _default_parquet_format, id="test_parquet_time64us"),
        pytest.param(pa.timestamp("s"), {"type": "string", "format": "date-time"}, _default_parquet_format, id="test_parquet_timestamps_s"),
        pytest.param(
            pa.timestamp("ms"), {"type": "string", "format": "date-time"}, _default_parquet_format, id="test_parquet_timestamp_ms"
        ),
        pytest.param(
            pa.timestamp("s", "utc"),
            {"type": "string", "format": "date-time"},
            _default_parquet_format,
            id="test_parquet_timestamps_s_with_tz",
        ),
        pytest.param(
            pa.timestamp("ms", "est"),
            {"type": "string", "format": "date-time"},
            _default_parquet_format,
            id="test_parquet_timestamps_ms_with_tz",
        ),
        pytest.param(pa.date32(), {"type": "string", "format": "date"}, _default_parquet_format, id="test_parquet_date32"),
        pytest.param(pa.date64(), {"type": "string", "format": "date"}, _default_parquet_format, id="test_parquet_date64"),
        pytest.param(pa.duration("s"), {"type": "integer"}, _default_parquet_format, id="test_duration_s"),
        pytest.param(pa.duration("ms"), {"type": "integer"}, _default_parquet_format, id="test_duration_ms"),
        pytest.param(pa.duration("us"), {"type": "integer"}, _default_parquet_format, id="test_duration_us"),
        pytest.param(pa.duration("ns"), {"type": "integer"}, _default_parquet_format, id="test_duration_ns"),
        pytest.param(pa.month_day_nano_interval(), {"type": "array"}, _default_parquet_format, id="test_parquet_month_day_nano_interval"),
        pytest.param(pa.binary(), {"type": "string"}, _default_parquet_format, id="test_binary"),
        pytest.param(pa.binary(2), {"type": "string"}, _default_parquet_format, id="test_fixed_size_binary"),
        pytest.param(pa.string(), {"type": "string"}, _default_parquet_format, id="test_parquet_string"),
        pytest.param(pa.utf8(), {"type": "string"}, _default_parquet_format, id="test_utf8"),
        pytest.param(pa.large_binary(), {"type": "string"}, _default_parquet_format, id="test_large_binary"),
        pytest.param(pa.large_string(), {"type": "string"}, _default_parquet_format, id="test_large_string"),
        pytest.param(pa.large_utf8(), {"type": "string"}, _default_parquet_format, id="test_large_utf8"),
        pytest.param(pa.dictionary(pa.int32(), pa.string()), {"type": "object"}, _default_parquet_format, id="test_dictionary"),
        pytest.param(pa.struct([pa.field("field", pa.int32())]), {"type": "object"}, _default_parquet_format, id="test_struct"),
        pytest.param(pa.list_(pa.int32()), {"type": "array"}, _default_parquet_format, id="test_list"),
        pytest.param(pa.large_list(pa.int32()), {"type": "array"}, _default_parquet_format, id="test_large_list"),
        pytest.param(pa.decimal128(2), {"type": "string"}, _default_parquet_format, id="test_decimal128"),
        pytest.param(pa.decimal256(2), {"type": "string"}, _default_parquet_format, id="test_decimal256"),
        pytest.param(pa.decimal128(2), {"type": "number"}, _decimal_as_float_parquet_format, id="test_decimal128_as_float"),
        pytest.param(pa.decimal256(2), {"type": "number"}, _decimal_as_float_parquet_format, id="test_decimal256_as_float"),
        pytest.param(pa.map_(pa.int32(), pa.int32()), {"type": "object"}, _default_parquet_format, id="test_map"),
        pytest.param(pa.null(), {"type": "null"}, _default_parquet_format, id="test_null"),
    ],
)
def test_type_mapping(parquet_type: pa.DataType, expected_type: Mapping[str, str], parquet_format: ParquetFormat) -> None:
    if expected_type is None:
        with pytest.raises(ValueError):
            ParquetParser.parquet_type_to_schema_type(parquet_type, parquet_format)
    else:
        assert ParquetParser.parquet_type_to_schema_type(parquet_type, parquet_format) == expected_type


@pytest.mark.parametrize(
    "pyarrow_type, parquet_format, parquet_object, expected_value",
    [
        pytest.param(pa.bool_(), _default_parquet_format, True, True, id="test_bool"),
        pytest.param(pa.int8(), _default_parquet_format, -1, -1, id="test_int8"),
        pytest.param(pa.int16(), _default_parquet_format, 2, 2, id="test_int16"),
        pytest.param(pa.int32(), _default_parquet_format, 3, 3, id="test_int32"),
        pytest.param(pa.int64(), _default_parquet_format, 4, 4, id="test_int64"),
        pytest.param(pa.uint8(), _default_parquet_format, 4, 4, id="test_parquet_uint8"),
        pytest.param(pa.uint16(), _default_parquet_format, 5, 5, id="test_parquet_uint16"),
        pytest.param(pa.uint32(), _default_parquet_format, 6, 6, id="test_parquet_uint32"),
        pytest.param(pa.uint64(), _default_parquet_format, 6, 6, id="test_parquet_uint64"),
        pytest.param(pa.float32(), _default_parquet_format, 2.7, 2.7, id="test_parquet_float32"),
        pytest.param(pa.float64(), _default_parquet_format, 3.14, 3.14, id="test_parquet_float64"),
        pytest.param(pa.time32("s"), _default_parquet_format, datetime.time(1, 2, 3), "01:02:03", id="test_parquet_time32s"),
        pytest.param(pa.time32("ms"), _default_parquet_format, datetime.time(3, 4, 5), "03:04:05", id="test_parquet_time32ms"),
        pytest.param(pa.time64("us"), _default_parquet_format, datetime.time(6, 7, 8), "06:07:08", id="test_parquet_time64us"),
        pytest.param(pa.time64("ns"), _default_parquet_format, datetime.time(9, 10, 11), "09:10:11", id="test_parquet_time64us"),
        pytest.param(
            pa.timestamp("s"),
            _default_parquet_format,
            datetime.datetime(2023, 7, 7, 10, 11, 12),
            "2023-07-07T10:11:12",
            id="test_parquet_timestamps_s",
        ),
        pytest.param(
            pa.timestamp("ms"),
            _default_parquet_format,
            datetime.datetime(2024, 8, 8, 11, 12, 13),
            "2024-08-08T11:12:13",
            id="test_parquet_timestamp_ms",
        ),
        pytest.param(
            pa.timestamp("s", "utc"),
            _default_parquet_format,
            datetime.datetime(2020, 1, 1, 1, 1, 1, tzinfo=datetime.timezone.utc),
            "2020-01-01T01:01:01+00:00",
            id="test_parquet_timestamps_s_with_tz",
        ),
        pytest.param(
            pa.timestamp("ms", "utc"),
            _default_parquet_format,
            datetime.datetime(2021, 2, 3, 4, 5, tzinfo=datetime.timezone.utc),
            "2021-02-03T04:05:00+00:00",
            id="test_parquet_timestamps_ms_with_tz",
        ),
        pytest.param(pa.date32(), _default_parquet_format, datetime.date(2023, 7, 7), "2023-07-07", id="test_parquet_date32"),
        pytest.param(pa.date64(), _default_parquet_format, datetime.date(2023, 7, 8), "2023-07-08", id="test_parquet_date64"),
        pytest.param(pa.duration("s"), _default_parquet_format, 12345, 12345, id="test_duration_s"),
        pytest.param(pa.duration("ms"), _default_parquet_format, 12345, 12345, id="test_duration_ms"),
        pytest.param(pa.duration("us"), _default_parquet_format, 12345, 12345, id="test_duration_us"),
        pytest.param(pa.duration("ns"), _default_parquet_format, 12345, 12345, id="test_duration_ns"),
        pytest.param(
            pa.month_day_nano_interval(),
            _default_parquet_format,
            datetime.timedelta(days=3, microseconds=4),
            [0, 3, 4000],
            id="test_parquet_month_day_nano_interval",
        ),
        pytest.param(pa.binary(), _default_parquet_format, b"this is a binary string", "this is a binary string", id="test_binary"),
        pytest.param(pa.binary(2), _default_parquet_format, b"t1", "t1", id="test_fixed_size_binary"),
        pytest.param(pa.string(), _default_parquet_format, "this is a string", "this is a string", id="test_parquet_string"),
        pytest.param(pa.utf8(), _default_parquet_format, "utf8".encode("utf8"), "utf8", id="test_utf8"),
        pytest.param(pa.large_binary(), _default_parquet_format, b"large binary string", "large binary string", id="test_large_binary"),
        pytest.param(pa.large_string(), _default_parquet_format, "large string", "large string", id="test_large_string"),
        pytest.param(pa.large_utf8(), _default_parquet_format, "large utf8", "large utf8", id="test_large_utf8"),
        pytest.param(pa.struct([pa.field("field", pa.int32())]), _default_parquet_format, {"field": 1}, {"field": 1}, id="test_struct"),
        pytest.param(pa.list_(pa.int32()), _default_parquet_format, [1, 2, 3], [1, 2, 3], id="test_list"),
        pytest.param(pa.large_list(pa.int32()), _default_parquet_format, [4, 5, 6], [4, 5, 6], id="test_large_list"),
        pytest.param(pa.decimal128(5, 3), _default_parquet_format, 12, "12.000", id="test_decimal128"),
        pytest.param(pa.decimal256(8, 2), _default_parquet_format, 13, "13.00", id="test_decimal256"),
        pytest.param(pa.decimal128(5, 3), _decimal_as_float_parquet_format, 12, 12.000, id="test_decimal128"),
        pytest.param(pa.decimal256(8, 2), _decimal_as_float_parquet_format, 13, 13.00, id="test_decimal256"),
        pytest.param(
            pa.map_(pa.string(), pa.int32()), _default_parquet_format, {"hello": 1, "world": 2}, {"hello": 1, "world": 2}, id="test_map"
        ),
        pytest.param(pa.null(), _default_parquet_format, None, None, id="test_null"),
    ],
)
def test_value_transformation(
    pyarrow_type: pa.DataType, parquet_format: ParquetFormat, parquet_object: Scalar, expected_value: Any
) -> None:
    pyarrow_value = pa.array([parquet_object], type=pyarrow_type)[0]
    py_value = ParquetParser._to_output_value(pyarrow_value, parquet_format)
    if isinstance(py_value, float):
        assert math.isclose(py_value, expected_value, abs_tol=0.01)
    else:
        assert py_value == expected_value


def test_value_dictionary() -> None:
    # Setting the dictionary is more involved than other data types so we test it in a separate test
    dictionary_values = ["apple", "banana", "cherry"]
    indices = [0, 1, 2, 0, 1]
    indices_array = pa.array(indices, type=pa.int8())
    dictionary = pa.DictionaryArray.from_arrays(indices_array, dictionary_values)
    py_value = ParquetParser._to_output_value(dictionary, _default_parquet_format)
    assert py_value == {"indices": [0, 1, 2, 0, 1], "values": ["apple", "banana", "cherry"]}


@pytest.mark.parametrize(
    "parquet_type, parquet_format",
    [
        pytest.param(pa.bool_(), _default_parquet_format, id="test_parquet_bool"),
        pytest.param(pa.int8(), _default_parquet_format, id="test_parquet_int8"),
        pytest.param(pa.int16(), _default_parquet_format, id="test_parquet_int16"),
        pytest.param(pa.int32(), _default_parquet_format, id="test_parquet_int32"),
        pytest.param(pa.int64(), _default_parquet_format, id="test_parquet_int64"),
        pytest.param(pa.uint8(), _default_parquet_format, id="test_parquet_uint8"),
        pytest.param(pa.uint16(), _default_parquet_format, id="test_parquet_uint16"),
        pytest.param(pa.uint32(), _default_parquet_format, id="test_parquet_uint32"),
        pytest.param(pa.uint64(), _default_parquet_format, id="test_parquet_uint64"),
        pytest.param(pa.float16(), _default_parquet_format, id="test_parquet_float16"),
        pytest.param(pa.float32(), _default_parquet_format, id="test_parquet_float32"),
        pytest.param(pa.float64(), _default_parquet_format, id="test_parquet_float64"),
        pytest.param(pa.time32("s"), _default_parquet_format, id="test_parquet_time32s"),
        pytest.param(pa.time32("ms"), _default_parquet_format, id="test_parquet_time32ms"),
        pytest.param(pa.time64("us"), _default_parquet_format, id="test_parquet_time64us"),
        pytest.param(pa.time64("ns"), _default_parquet_format, id="test_parquet_time64ns"),
        pytest.param(pa.timestamp("s"), _default_parquet_format, id="test_parquet_timestamps_s"),
        pytest.param(pa.timestamp("ms"), _default_parquet_format, id="test_parquet_timestamp_ms"),
        pytest.param(pa.timestamp("s", "utc"), _default_parquet_format, id="test_parquet_timestamps_s_with_tz"),
        pytest.param(pa.timestamp("ms", "est"), _default_parquet_format, id="test_parquet_timestamps_ms_with_tz"),
        pytest.param(pa.date32(), _default_parquet_format, id="test_parquet_date32"),
        pytest.param(pa.date64(), _default_parquet_format, id="test_parquet_date64"),
        pytest.param(pa.duration("s"), _default_parquet_format, id="test_duration_s"),
        pytest.param(pa.duration("ms"), _default_parquet_format, id="test_duration_ms"),
        pytest.param(pa.duration("us"), _default_parquet_format, id="test_duration_us"),
        pytest.param(pa.duration("ns"), _default_parquet_format, id="test_duration_ns"),
        pytest.param(pa.month_day_nano_interval(), _default_parquet_format, id="test_parquet_month_day_nano_interval"),
        pytest.param(pa.binary(), _default_parquet_format, id="test_binary"),
        pytest.param(pa.binary(2), _default_parquet_format, id="test_fixed_size_binary"),
        pytest.param(pa.string(), _default_parquet_format, id="test_parquet_string"),
        pytest.param(pa.utf8(), _default_parquet_format, id="test_utf8"),
        pytest.param(pa.large_binary(), _default_parquet_format, id="test_large_binary"),
        pytest.param(pa.large_string(), _default_parquet_format, id="test_large_string"),
        pytest.param(pa.large_utf8(), _default_parquet_format, id="test_large_utf8"),
        pytest.param(pa.dictionary(pa.int32(), pa.string()), _default_parquet_format, id="test_dictionary"),
        pytest.param(pa.struct([pa.field("field", pa.int32())]), _default_parquet_format, id="test_struct"),
        pytest.param(pa.list_(pa.int32()), _default_parquet_format, id="test_list"),
        pytest.param(pa.large_list(pa.int32()), _default_parquet_format, id="test_large_list"),
        pytest.param(pa.decimal128(2), _default_parquet_format, id="test_decimal128"),
        pytest.param(pa.decimal256(2), _default_parquet_format, id="test_decimal256"),
        pytest.param(pa.decimal128(2), _decimal_as_float_parquet_format, id="test_decimal128_as_float"),
        pytest.param(pa.decimal256(2), _decimal_as_float_parquet_format, id="test_decimal256_as_float"),
        pytest.param(pa.map_(pa.int32(), pa.int32()), _default_parquet_format, id="test_map"),
        pytest.param(pa.null(), _default_parquet_format, id="test_null"),
    ])
def test_null_value_does_not_throw(parquet_type, parquet_format) -> None:
    pyarrow_value = pa.scalar(None, type=parquet_type)
    assert ParquetParser._to_output_value(pyarrow_value, parquet_format) is None


@pytest.mark.parametrize(
    "file_format",
    [
        pytest.param(
            CsvFormat(
                filetype="csv",
                delimiter=",",
                escape_char="\\",
                quote_char='"',
            ),
            id="test_csv_format",
        ),
        pytest.param(JsonlFormat(), id="test_jsonl_format"),
    ],
)
def test_wrong_file_format(file_format: Union[CsvFormat, JsonlFormat]) -> None:
    parser = ParquetParser()
    config = FileBasedStreamConfig(
        name="test.parquet",
        file_type=file_format.filetype,
        format={file_format.filetype: file_format},
        validation_policy=ValidationPolicy.emit_record,
    )
    file = RemoteFile(uri="s3://mybucket/test.parquet", last_modified=datetime.datetime.now())
    stream_reader = Mock()
    logger = Mock()
    with pytest.raises(ValueError):
        asyncio.get_event_loop().run_until_complete(parser.infer_schema(config, file, stream_reader, logger))



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import logging
from typing import Dict, Iterable, List, Optional, Set

from airbyte_cdk.models import AirbyteMessage, AirbyteStreamStatus
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.concurrent_source.partition_generation_completed_sentinel import PartitionGenerationCompletedSentinel
from airbyte_cdk.sources.concurrent_source.thread_pool_manager import ThreadPoolManager
from airbyte_cdk.sources.message import MessageRepository
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.partition_enqueuer import PartitionEnqueuer
from airbyte_cdk.sources.streams.concurrent.partition_reader import PartitionReader
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.record import Record
from airbyte_cdk.sources.streams.concurrent.partitions.types import PartitionCompleteSentinel
from airbyte_cdk.sources.utils.record_helper import stream_data_to_airbyte_message
from airbyte_cdk.sources.utils.slice_logger import SliceLogger
from airbyte_cdk.utils.stream_status_utils import as_airbyte_message as stream_status_as_airbyte_message


class ConcurrentReadProcessor:
    def __init__(
        self,
        stream_instances_to_read_from: List[AbstractStream],
        partition_enqueuer: PartitionEnqueuer,
        thread_pool_manager: ThreadPoolManager,
        logger: logging.Logger,
        slice_logger: SliceLogger,
        message_repository: MessageRepository,
        partition_reader: PartitionReader,
    ):
        """
        This class is responsible for handling items from a concurrent stream read process.
        :param stream_instances_to_read_from: List of streams to read from
        :param partition_enqueuer: PartitionEnqueuer instance
        :param thread_pool_manager: ThreadPoolManager instance
        :param logger: Logger instance
        :param slice_logger: SliceLogger instance
        :param message_repository: MessageRepository instance
        :param partition_reader: PartitionReader instance
        """
        self._stream_name_to_instance = {s.name: s for s in stream_instances_to_read_from}
        self._record_counter = {}
        self._streams_to_running_partitions: Dict[str, Set[Partition]] = {}
        for stream in stream_instances_to_read_from:
            self._streams_to_running_partitions[stream.name] = set()
            self._record_counter[stream.name] = 0
        self._thread_pool_manager = thread_pool_manager
        self._partition_enqueuer = partition_enqueuer
        self._stream_instances_to_start_partition_generation = stream_instances_to_read_from
        self._streams_currently_generating_partitions: List[str] = []
        self._logger = logger
        self._slice_logger = slice_logger
        self._message_repository = message_repository
        self._partition_reader = partition_reader
        self._streams_done: Set[str] = set()

    def on_partition_generation_completed(self, sentinel: PartitionGenerationCompletedSentinel) -> Iterable[AirbyteMessage]:
        """
        This method is called when a partition generation is completed.
        1. Remove the stream from the list of streams currently generating partitions
        2. If the stream is done, mark it as such and return a stream status message
        3. If there are more streams to read from, start the next partition generator
        """
        stream_name = sentinel.stream.name
        self._streams_currently_generating_partitions.remove(sentinel.stream.name)
        # It is possible for the stream to already be done if no partitions were generated
        # If the partition generation process was completed and there are no partitions left to process, the stream is done
        if self._is_stream_done(stream_name) or len(self._streams_to_running_partitions[stream_name]) == 0:
            yield from self._on_stream_is_done(stream_name)
        if self._stream_instances_to_start_partition_generation:
            yield self.start_next_partition_generator()

    def on_partition(self, partition: Partition) -> None:
        """
        This method is called when a partition is generated.
        1. Add the partition to the set of partitions for the stream
        2. Log the slice if necessary
        3. Submit the partition to the thread pool manager
        """
        stream_name = partition.stream_name()
        self._streams_to_running_partitions[stream_name].add(partition)
        if self._slice_logger.should_log_slice_message(self._logger):
            self._message_repository.emit_message(self._slice_logger.create_slice_log_message(partition.to_slice()))
        self._thread_pool_manager.submit(self._partition_reader.process_partition, partition)

    def on_partition_complete_sentinel(self, sentinel: PartitionCompleteSentinel) -> Iterable[AirbyteMessage]:
        """
        This method is called when a partition is completed.
        1. Close the partition
        2. If the stream is done, mark it as such and return a stream status message
        3. Emit messages that were added to the message repository
        """
        partition = sentinel.partition
        partition.close()
        partitions_running = self._streams_to_running_partitions[partition.stream_name()]
        if partition in partitions_running:
            partitions_running.remove(partition)
            # If all partitions were generated and this was the last one, the stream is done
            if partition.stream_name() not in self._streams_currently_generating_partitions and len(partitions_running) == 0:
                yield from self._on_stream_is_done(partition.stream_name())
        yield from self._message_repository.consume_queue()

    def on_record(self, record: Record) -> Iterable[AirbyteMessage]:
        """
        This method is called when a record is read from a partition.
        1. Convert the record to an AirbyteMessage
        2. If this is the first record for the stream, mark the stream as RUNNING
        3. Increment the record counter for the stream
        4. Emit the message
        5. Emit messages that were added to the message repository
        """
        # Do not pass a transformer or a schema
        # AbstractStreams are expected to return data as they are expected.
        # Any transformation on the data should be done before reaching this point
        message = stream_data_to_airbyte_message(record.stream_name, record.data)
        stream = self._stream_name_to_instance[record.stream_name]

        if message.type == MessageType.RECORD:
            if self._record_counter[stream.name] == 0:
                self._logger.info(f"Marking stream {stream.name} as RUNNING")
                yield stream_status_as_airbyte_message(stream.as_airbyte_stream(), AirbyteStreamStatus.RUNNING)
            self._record_counter[stream.name] += 1
        yield message
        yield from self._message_repository.consume_queue()

    def on_exception(self, exception: Exception) -> Iterable[AirbyteMessage]:
        """
        This method is called when an exception is raised.
        1. Stop all running streams
        2. Raise the exception
        """
        yield from self._stop_streams()
        raise exception

    def start_next_partition_generator(self) -> Optional[AirbyteMessage]:
        """
        Start the next partition generator.
        1. Pop the next stream to read from
        2. Submit the partition generator to the thread pool manager
        3. Add the stream to the list of streams currently generating partitions
        4. Return a stream status message
        """
        if self._stream_instances_to_start_partition_generation:
            stream = self._stream_instances_to_start_partition_generation.pop(0)
            self._thread_pool_manager.submit(self._partition_enqueuer.generate_partitions, stream)
            self._streams_currently_generating_partitions.append(stream.name)
            self._logger.info(f"Marking stream {stream.name} as STARTED")
            self._logger.info(f"Syncing stream: {stream.name} ")
            return stream_status_as_airbyte_message(
                stream.as_airbyte_stream(),
                AirbyteStreamStatus.STARTED,
            )
        else:
            return None

    def is_done(self) -> bool:
        """
        This method is called to check if the sync is done.
        The sync is done when:
        1. There are no more streams generating partitions
        2. There are no more streams to read from
        3. All partitions for all streams are closed
        """
        return all([self._is_stream_done(stream_name) for stream_name in self._stream_name_to_instance.keys()])

    def _is_stream_done(self, stream_name: str) -> bool:
        return stream_name in self._streams_done

    def _on_stream_is_done(self, stream_name: str) -> Iterable[AirbyteMessage]:
        self._logger.info(f"Read {self._record_counter[stream_name]} records from {stream_name} stream")
        self._logger.info(f"Marking stream {stream_name} as STOPPED")
        stream = self._stream_name_to_instance[stream_name]
        stream.cursor.ensure_at_least_one_state_emitted()
        yield from self._message_repository.consume_queue()
        self._logger.info(f"Finished syncing {stream.name}")
        self._streams_done.add(stream_name)
        yield stream_status_as_airbyte_message(stream.as_airbyte_stream(), AirbyteStreamStatus.COMPLETE)

    def _stop_streams(self) -> Iterable[AirbyteMessage]:
        self._thread_pool_manager.shutdown()
        for stream_name in self._streams_to_running_partitions.keys():
            stream = self._stream_name_to_instance[stream_name]
            if not self._is_stream_done(stream_name):
                self._logger.info(f"Marking stream {stream.name} as STOPPED")
                self._logger.info(f"Finished syncing {stream.name}")
                yield stream_status_as_airbyte_message(stream.as_airbyte_stream(), AirbyteStreamStatus.INCOMPLETE)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Any, List, MutableMapping, Tuple

if TYPE_CHECKING:
    from airbyte_cdk.sources.streams.concurrent.cursor import CursorField


class ConcurrencyCompatibleStateType(Enum):
    date_range = "date-range"


class AbstractStreamStateConverter(ABC):
    START_KEY = "start"
    END_KEY = "end"

    @abstractmethod
    def deserialize(self, state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        """
        Perform any transformations needed for compatibility with the converter.
        """
        ...

    @staticmethod
    def is_state_message_compatible(state: MutableMapping[str, Any]) -> bool:
        return bool(state) and state.get("state_type") in [t.value for t in ConcurrencyCompatibleStateType]

    @abstractmethod
    def convert_from_sequential_state(
        self,
        cursor_field: "CursorField",
        stream_state: MutableMapping[str, Any],
        start: Any,
    ) -> Tuple[Any, MutableMapping[str, Any]]:
        """
        Convert the state message to the format required by the ConcurrentCursor.

        e.g.
        {
            "state_type": ConcurrencyCompatibleStateType.date_range.value,
            "metadata": { … },
            "slices": [
                {starts: 0, end: 1617030403, finished_processing: true}]
        }
        """
        ...

    @abstractmethod
    def convert_to_sequential_state(self, cursor_field: "CursorField", stream_state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        """
        Convert the state message from the concurrency-compatible format to the stream's original format.

        e.g.
        { "created": 1617030403 }
        """
        ...

    @abstractmethod
    def increment(self, timestamp: Any) -> Any:
        """
        Increment a timestamp by a single unit.
        """
        ...

    @abstractmethod
    def merge_intervals(self, intervals: List[MutableMapping[str, Any]]) -> List[MutableMapping[str, Any]]:
        """
        Compute and return a list of merged intervals.

        Intervals may be merged if the start time of the second interval is 1 unit or less (as defined by the
        `increment` method) than the end time of the first interval.
        """
        ...

    @abstractmethod
    def parse_value(self, value: Any) -> Any:
        """
        Parse the value of the cursor field into a comparable value.
        """
        ...

    @property
    @abstractmethod
    def zero_value(self) -> Any:
        ...



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

from airbyte_cdk.test.mock_http.request import HttpRequest


class HttpRequestMatcher:
    def __init__(self, request: HttpRequest, minimum_number_of_expected_match: int):
        self._request_to_match = request
        self._minimum_number_of_expected_match = minimum_number_of_expected_match
        self._actual_number_of_matches = 0

    def matches(self, request: HttpRequest) -> bool:
        hit = request.matches(self._request_to_match)
        if hit:
            self._actual_number_of_matches += 1
        return hit

    def has_expected_match_count(self) -> bool:
        return self._actual_number_of_matches >= self._minimum_number_of_expected_match

    @property
    def actual_number_of_matches(self) -> int:
        return self._actual_number_of_matches

    @property
    def request(self) -> HttpRequest:
        return self._request_to_match

    def __str__(self) -> str:
        return (
            f"HttpRequestMatcher("
            f"request_to_match={self._request_to_match}, "
            f"minimum_number_of_expected_match={self._minimum_number_of_expected_match}, "
            f"actual_number_of_matches={self._actual_number_of_matches})"
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import dataclasses
from datetime import datetime
from typing import Any, Mapping

from airbyte_cdk.connector_builder.message_grouper import MessageGrouper
from airbyte_cdk.models import AirbyteMessage, AirbyteRecordMessage, ConfiguredAirbyteCatalog
from airbyte_cdk.models import Type
from airbyte_cdk.models import Type as MessageType
from airbyte_cdk.sources.declarative.declarative_source import DeclarativeSource
from airbyte_cdk.sources.declarative.manifest_declarative_source import ManifestDeclarativeSource
from airbyte_cdk.sources.declarative.parsers.model_to_component_factory import ModelToComponentFactory
from airbyte_cdk.utils.traced_exception import AirbyteTracedException

DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE = 5
DEFAULT_MAXIMUM_NUMBER_OF_SLICES = 5
DEFAULT_MAXIMUM_RECORDS = 100

MAX_PAGES_PER_SLICE_KEY = "max_pages_per_slice"
MAX_SLICES_KEY = "max_slices"
MAX_RECORDS_KEY = "max_records"


@dataclasses.dataclass
class TestReadLimits:
    max_records: int = dataclasses.field(default=DEFAULT_MAXIMUM_RECORDS)
    max_pages_per_slice: int = dataclasses.field(default=DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE)
    max_slices: int = dataclasses.field(default=DEFAULT_MAXIMUM_NUMBER_OF_SLICES)


def get_limits(config: Mapping[str, Any]) -> TestReadLimits:
    command_config = config.get("__test_read_config", {})
    max_pages_per_slice = command_config.get(MAX_PAGES_PER_SLICE_KEY) or DEFAULT_MAXIMUM_NUMBER_OF_PAGES_PER_SLICE
    max_slices = command_config.get(MAX_SLICES_KEY) or DEFAULT_MAXIMUM_NUMBER_OF_SLICES
    max_records = command_config.get(MAX_RECORDS_KEY) or DEFAULT_MAXIMUM_RECORDS
    return TestReadLimits(max_records, max_pages_per_slice, max_slices)


def create_source(config: Mapping[str, Any], limits: TestReadLimits) -> ManifestDeclarativeSource:
    manifest = config["__injected_declarative_manifest"]
    return ManifestDeclarativeSource(
        emit_connector_builder_messages=True,
        source_config=manifest,
        component_factory=ModelToComponentFactory(
            emit_connector_builder_messages=True,
            limit_pages_fetched_per_slice=limits.max_pages_per_slice,
            limit_slices_fetched=limits.max_slices,
            disable_retries=True,
        ),
    )


def read_stream(
    source: DeclarativeSource, config: Mapping[str, Any], configured_catalog: ConfiguredAirbyteCatalog, limits: TestReadLimits
) -> AirbyteMessage:
    try:
        handler = MessageGrouper(limits.max_pages_per_slice, limits.max_slices, limits.max_records)
        stream_name = configured_catalog.streams[0].stream.name  # The connector builder only supports a single stream
        stream_read = handler.get_message_groups(source, config, configured_catalog, limits.max_records)
        return AirbyteMessage(
            type=MessageType.RECORD,
            record=AirbyteRecordMessage(data=dataclasses.asdict(stream_read), stream=stream_name, emitted_at=_emitted_at()),
        )
    except Exception as exc:
        error = AirbyteTracedException.from_exception(
            exc, message=f"Error reading stream with config={config} and catalog={configured_catalog}: {str(exc)}"
        )
        return error.as_airbyte_message()


def resolve_manifest(source: ManifestDeclarativeSource) -> AirbyteMessage:
    try:
        return AirbyteMessage(
            type=Type.RECORD,
            record=AirbyteRecordMessage(
                data={"manifest": source.resolved_manifest},
                emitted_at=_emitted_at(),
                stream="resolve_manifest",
            ),
        )
    except Exception as exc:
        error = AirbyteTracedException.from_exception(exc, message=f"Error resolving manifest: {str(exc)}")
        return error.as_airbyte_message()


def _emitted_at() -> int:
    return int(datetime.now().timestamp()) * 1000
